mod helper;
pub mod layout;

use cdtypes::types::{cdstring::DString, dir_record::DirectoryRecord, helper::{round_bytes_mode2_form1, sector_count_mode2_form1}, sector::*, path_table::PathTableL};
use layout::Layout;
use tool_helper::Error;
use std::{cell::RefCell, rc::Rc};

pub type SharedPtr<T> = Rc<RefCell<T>>;

pub fn new_shared_ptr<T>(value: T) -> SharedPtr<T> {
    Rc::new(RefCell::new(value))
}

pub struct CDDesc {
    system_area: SharedPtr<SystemArea>,
    pvd:         SharedPtr<PrimaryVolumeDescriptor>,
    root:        SharedPtr<Directory>
}

impl CDDesc {
    const DEFAULT_DATA_SIZE_FOR_NOW:usize = Mode2Form1::DATA_SIZE;

    pub fn new() -> CDDesc {
        match Directory::new("root") {
            Ok(root)   => CDDesc{system_area: new_shared_ptr(SystemArea::new()), pvd: new_shared_ptr(PrimaryVolumeDescriptor::new()), root: new_shared_ptr(root)},
            Err(error) => panic!("Creating root directory failed with: {}", error)
        }
    }

    pub fn get_memory_layout(&self) -> layout::MemoryLayout {
        layout::DefaultLayout::new(self)
    }

    pub fn add_dir(&mut self, dir: Directory) {
        self.root.borrow_mut().add_dir(dir);
    }

    pub fn add_file(&mut self, file: File) {
        self.root.borrow_mut().add_file(file);
    }

    pub fn calculate_lbas(&mut self) {
        let mut path_table_properties = {
            let mut size_bytes = 0;

            helper::collect_path_table_member(self.root.clone()).into_iter().for_each(|element| {
                println!("PT: {} ^{}", element.name, element.parent_table_id);
                size_bytes += PathTableL::calculate_size_for(element.name.as_ref());
            });

            size_bytes = round_bytes_mode2_form1(size_bytes)*4;
            Properties{lba: 0, overwrite_size_bytes: Some(size_bytes), is_hidden: false}
        };
        let mut cur_lba = 0;    

        Self::for_each_dir_mut(self.root.clone(), &|dir| {dir.update_content_size();});

        // Now layout iterate?
        for element in self.get_memory_layout() {
            fn update_lba(properties: &mut Properties, cur_lba: usize, content_size: usize) -> usize {
                properties.lba = cur_lba;
                cur_lba + properties.sector_count_xa_data(content_size)
            }
            
            match element {
                Layout::SystemArea(system_area) => {
                    let mut system_area = system_area.borrow_mut();
                    cur_lba = update_lba(&mut system_area.properties, cur_lba, Self::DEFAULT_DATA_SIZE_FOR_NOW);
                },
                Layout::PVD(pvd)                => {
                    let mut pvd = pvd.borrow_mut();
                    cur_lba = update_lba(&mut pvd.properties, cur_lba, Self::DEFAULT_DATA_SIZE_FOR_NOW);
                },
                Layout::PathTables              => {
                    cur_lba = update_lba(&mut path_table_properties, cur_lba, Self::DEFAULT_DATA_SIZE_FOR_NOW);
                },
                Layout::Directory(dir) => {
                    let mut dir = dir.borrow_mut();
                    cur_lba = update_lba(&mut dir.properties, cur_lba, Self::DEFAULT_DATA_SIZE_FOR_NOW)
                },
                Layout::File(file)                     => {
                    let mut file = file.borrow_mut();
                    cur_lba = update_lba(&mut file.properties, cur_lba, Self::DEFAULT_DATA_SIZE_FOR_NOW);
                }
            }

            println!("cur_lba: {}", cur_lba);
        }
    }

    fn for_each_dir_mut<T: Fn(&mut Directory)>(dir: SharedPtr<Directory>, function: &T) {
        let mut dir = dir.borrow_mut();

        function(&mut dir);
        for sub_dir in &mut dir.dirs {
            Self::for_each_dir_mut(sub_dir.clone(), function);
        }
    }
}

pub struct SystemArea {
    pub properties: Properties
}

impl SystemArea {
    const DEFAULT_SIZE:usize = (16*Mode2Form1::DATA_SIZE);

    pub fn new() -> SystemArea {
        SystemArea{properties: Properties{lba: 0, overwrite_size_bytes: Some(Self::DEFAULT_SIZE), is_hidden: false}}
    }
}

pub struct PrimaryVolumeDescriptor {
    pub properties: Properties
}

impl PrimaryVolumeDescriptor {
    const DEFAULT_SIZE:usize = (2*Mode2Form1::DATA_SIZE);

    pub fn new() -> PrimaryVolumeDescriptor {
        PrimaryVolumeDescriptor{properties: Properties{lba: 0, overwrite_size_bytes: Some(Self::DEFAULT_SIZE), is_hidden: false}}
    }
}

pub struct Directory {
    pub name:       DirectoryName,
    pub properties: Properties,
        files:      Vec<SharedPtr<File>>,
        dirs:       Vec<SharedPtr<Directory>>
}

impl Directory {
    pub fn new(dir_name: &str) -> Result<Directory, Error> {
        Ok(Directory{name: DirectoryName::from_str(dir_name)?, properties: Properties::default(), files: Vec::new(), dirs: Vec::new()})
    }

    pub fn add_dir(&mut self, dir: Directory) {
        self.dirs.push(new_shared_ptr(dir));
    }

    pub fn add_file(&mut self, file: File) {
        self.files.push(new_shared_ptr(file));
    }

    fn update_content_size(&mut self) {
        let dir_member     = helper::collect_directory_record_member(&self.files, &self.dirs);
        let mut size_bytes = 0;

        for member in dir_member {
            size_bytes += DirectoryRecord::calculate_size_for(member.get_name().as_str(), true);
        }

        self.properties.overwrite_size_bytes = Some(size_bytes);
    }  

    fn _update_content(&mut self) {
        let dir_member = helper::collect_directory_record_member(&self.files, &self.dirs);

        println!("{} updating content", self.name.as_string().unwrap_or("<No name>".to_owned()));
        for member in dir_member {
            println!(">>> {}", member.get_name());
        }        
    }
}

impl std::fmt::Display for Directory {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Directory: {}", self.name)
    }    
}

pub struct File {
    pub name:       FileName,
    pub properties: Properties,
        _content:    Vec<u8>
}

impl File {
    pub fn new(file_name: &str) -> Result<File, Error> {
        Ok(File{name: FileName::from_str(file_name)?, properties: Properties::default(), _content: Vec::new()})
    }
}

impl std::fmt::Display for File {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "File: {}", self.name)
    }
}

pub struct DirectoryName {
    name: DString<8>,
    len:  usize,
}

impl DirectoryName {
    pub fn from_str(dir_name: &str) -> Result<DirectoryName, Error> {
        let dir_name = dir_name.to_uppercase();
        Ok(DirectoryName{name: DString::from_str(dir_name.as_ref())?, len: dir_name.len()})
    }

    pub fn as_string(&self) -> Option<String> {
        Some(self.as_str()?.to_owned())
    }

    fn as_str(&self) -> Option<&str> {
        dstring_as_str(&self.name, self.len)
    }
}

impl std::fmt::Display for DirectoryName {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.as_str().unwrap_or("<???>"))
    } 
}

pub struct FileName {
    name:    DString<8>,
    len:     usize,
    ext:     Option<DString<3>>,
    ext_len: Option<usize>
}

impl FileName {
    pub fn from_str(file_name: &str) -> Result<FileName, Error> {
        let file_name = file_name.to_uppercase();
        let (name, ext) = {
            let mut sub_str = file_name.split('.');
            (Error::ok_or_new(sub_str.next(), || "File name can't be emplty".to_owned())?, sub_str.next())
        };
        let (ext, ext_len) = {
            match ext {
                Some(ext) => (Some(DString::from_str(ext)?), Some(ext.len())),
                None      => (None, None)
            }   
        };

        Ok(FileName{name: DString::from_str(name)?, len: name.len(), ext, ext_len})
    }

    pub fn as_string(&self) -> Option<String> {
        if let Some(ext) = &self.ext {
            Some(format!("{}.{};1", dstring_as_str(&self.name, self.len)?, dstring_as_str(&ext, self.ext_len.unwrap_or(0))?))
        }

        else {
            Some(format!("{}", dstring_as_str(&self.name, self.len)?))
        }
    }
}

impl std::fmt::Display for FileName {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.as_string().unwrap_or("<???>".to_owned()))
    }
}

pub struct Properties {
    pub lba:                  usize,
    pub overwrite_size_bytes: Option<usize>,
    pub is_hidden:            bool
}

impl Properties {
    pub fn sector_count_xa_data(&self, content_size: usize) -> usize {
        let size_bytes = { 
            if let Some(forced_bytes) = self.overwrite_size_bytes {
                forced_bytes
            }

            else {
                content_size
            }
        };

        sector_count_mode2_form1(size_bytes)
    }
}

impl Default for Properties {
    fn default() -> Self {
        Properties{lba: 0, overwrite_size_bytes: None, is_hidden: false}
    }
}

fn dstring_as_str<const SIZE: usize>(string: &DString<SIZE>, len: usize) -> Option<&str> {
    match std::str::from_utf8(&string.as_raw()[0..len]) {
        Ok(str) => Some(str),
        Err(_)  => None,
    }
}